/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "via_driver.h"
#include "via_regs.h"
#include "via_rotate.h"

union NewFloat
{
    float   OriginalType;
    CARD32  IntegerType;
};


/* Get reciprocal of source pitch. ex: 1/sp = 0.x, if sp=256d, x=004000h */ 
unsigned int GetReciprocal(int dwSourcePitch)
{    
    CARD32  dwexp = 0;
    CARD32  dwtep = 0;
    CARD32  dwRecptial = 0;
    float   frecptial = 0000.0f;               
    union NewFloat TmpFloat;
    
    frecptial = (float)dwSourcePitch;
    frecptial = 1.0f/frecptial;
    
    TmpFloat.OriginalType = frecptial;
    dwRecptial = TmpFloat.IntegerType;               

    /* Get 23 bits Mantissa */     
    dwtep = dwRecptial & 0x7FFFFF;               
    
    /*  Default bit 23 is one 0-23, total 24 bits   */
    dwtep |= 0x800000;
    
    /*  shift bits 23 of Mantissa to bit 31 (23 -0) to (31 - 8) */
    dwtep <<= 8;        
    
    /* Get 8 bits exp */    
    dwexp = (dwRecptial & 0x7F800000)>>23;
    
    /* Sub Bias 127 */
    dwexp -= 127;
    
    /* dwexp always be negative becausse the range of result is 1 - 0  */
    dwexp *= -1;

    /* the virtual value "1" had be shift to bit 31 (the first bit of Mantissa)  */
    dwexp--;       
     
    /* Shift Mantissa according dwexp   */    
    dwtep >>= dwexp;    
    
    /* Shift bit 31 of Mantissa to bit 21 */
    /* get 0-21, total 22 bits Mantissa*/
    dwtep >>= 10;        
    
    return (CARD32)dwtep;
}



/* Enable direct access window to rotate the screen. */
void EnableDirectAccessWindow(ScrnInfoPtr pScrn)
{    
    VIAPtr  pVia = VIAPTR(pScrn);    
    VIABIOSInfoPtr  pBIOSInfo;
    pBIOSInfo = pVia->pBIOSInfo;
    
    int     dwRotateDegree = pBIOSInfo->HWRotate;    
    CARD32  dwWidth = pBIOSInfo->VirtualX;  /* This value is before rotate. */
    CARD32  dwHeight = pBIOSInfo->VirtualY; /* This value is before rotate. */
    CARD32  dwWidthAlignment8 = 0;
    
    CARD32 dw1E00 = 0;    /* Degree, Bpp, Enable      */
    CARD32 dw1E04 = 0;    /* start address            */
    CARD32 dw1E08 = 0;    /* end address              */
    CARD32 dw1E0C = 0;    /* source pitch             */
    CARD32 dw1E10 = 0;    /* rotate pitch             */
    CARD32 dw1E14 = 0;    /* W & H                    */    

    /* To do alignment, like 1366x768, 1400x1050... */
    dwWidthAlignment8 = (dwWidth + 7) & (~7);     

    if ((dwWidth == 1400) && (dwHeight == 1050) &&
        (pScrn->bitsPerPixel == 8) &&
        ((dwRotateDegree == VIA_ROTATE_DEGREE_90) || (dwRotateDegree == VIA_ROTATE_DEGREE_270)))
    {
        dwHeight += 2;
    }

    /* Set Degree */
    switch (dwRotateDegree)
    {
        case VIA_ROTATE_DEGREE_90:
            dw1E00 |= 0x8;
            break;
            
        case VIA_ROTATE_DEGREE_180:
            dw1E00 |= 0x10;
            break;
                    
        case VIA_ROTATE_DEGREE_270:
            dw1E00 |= 0x18;
            break;
    }

    /* Set Bpp */
    switch (pScrn->bitsPerPixel)
    {
        case 8:
            dw1E00 |= 0x0;
            break;
            
        case 16:
            dw1E00 |= 0x2;
            break;
                    
        case 32:
            dw1E00 |= 0x4;
            break;
    }

    /* 1E08, Set Rotate End Address */
    dw1E08 = dw1E04 + (dwWidthAlignment8 * dwHeight * (pScrn->bitsPerPixel/8));
            
    /* 1E0C, Rotate Source Pitch */
    /* 1E10, Rotate Pitch */
    /* 1E14, Rotate Height and Width */
    switch (dwRotateDegree)
    {
        case VIA_ROTATE_DEGREE_90:
        case VIA_ROTATE_DEGREE_270:
            dw1E0C = GetReciprocal(dwHeight); 
            dw1E10 = (dwWidthAlignment8 << 16) | dwHeight; 
            dw1E14 = (dwWidthAlignment8 << 16) | dwHeight;
            break;

        case VIA_ROTATE_DEGREE_0:
        case VIA_ROTATE_DEGREE_180:
            dw1E0C = GetReciprocal(dwWidthAlignment8);
            dw1E10 = (dwWidthAlignment8 << 16) | dwWidthAlignment8;
            dw1E14 = (dwHeight << 16) | dwWidthAlignment8;
            break;
    }        
    
    /* Set Registers */       
    VIASETREG(0x1E04, dw1E04);
    VIASETREG(0x1E08, dw1E08);
    VIASETREG(0x1E0C, dw1E0C);
    VIASETREG(0x1E10, dw1E10);
    VIASETREG(0x1E14, dw1E14);        
    VIASETREG(0x1E00, dw1E00 | 0x1); /* Fire */

}


void DisableDirectAccessWindow(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn); 
    
    /* Clear */       
    VIASETREG(0x1E04, 0);
    VIASETREG(0x1E08, 0);
    VIASETREG(0x1E0C, 0);
    VIASETREG(0x1E10, 0);
    VIASETREG(0x1E14, 0);        
    VIASETREG(0x1E00, 0);
}



/* Enable 2D Engine rotate function to acceleration 2D function. */
void Enable2DEngineRotate(ScrnInfoPtr pScrn, CARD32* pGEMode, CARD32* pCommand)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo;
    pBIOSInfo = pVia->pBIOSInfo;  
    CARD32  dwWidth = 0;
    CARD32  dwHeight = 0;
    
    /* we need to get the actual screen size but not the mode we set */
    dwWidth = pBIOSInfo->VirtualX;
    dwHeight = pBIOSInfo->VirtualY;
    
    CARD32  RotateX = 0, RotateY = 0, RotateDegree = 0; /* For rotate used. */
     
    dwWidth = (dwWidth + 7) & ~7;

    if ((pScrn->virtualX == 1400) && (pScrn->virtualY == 1050) &&
        (pScrn->bitsPerPixel == 8) &&
        ((pBIOSInfo->HWRotate == VIA_ROTATE_DEGREE_90) ||
         (pBIOSInfo->HWRotate == VIA_ROTATE_DEGREE_270)))
    {
        dwHeight += 2;
    }
    
    switch (pBIOSInfo->HWRotate)
    {
        case VIA_ROTATE_DEGREE_90:
            RotateX = dwHeight;
            RotateY = dwWidth;
            RotateDegree = VIA_GEM_ROTATE_90;            
            break;
            
        case VIA_ROTATE_DEGREE_180:
            RotateX = dwWidth;
            RotateY = dwHeight;
            RotateDegree = VIA_GEM_ROTATE_180;
            break;
            
        case VIA_ROTATE_DEGREE_270:
            RotateX = dwHeight;
            RotateY = dwWidth;
            RotateDegree = VIA_GEM_ROTATE_270;            
            break;                    
    }

    if (pVia->IsHWRotateEnabled)
    {
        /* Modify GEMode and Command according to rotate degree. */
        *pGEMode |= RotateDegree;
        *pCommand |= VIA_GEC_ROT;

        /* Set resolution of rotation surface. */
        VIASETREG(VIA_REG_ROTSRC, (((RotateY-1)<<16))|(RotateX-1) );
        VIASETREG(VIA_REG_ROTDST, (((RotateY-1)<<16))|(RotateX-1) );        
    }
}


/* Enable 2D Engine rotate function to acceleration 2D function. */
void Enable2DEngineRotate_M1(ScrnInfoPtr pScrn, CARD32* pGEMode, CARD32* pCommand)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo;
    pBIOSInfo = pVia->pBIOSInfo;    

    CARD32 RotateDegree = 0;
   
    switch (pBIOSInfo->HWRotate)
    {
        case VIA_ROTATE_DEGREE_90:             
            RotateDegree = VIA_GEM_ROTATE_90;            
            break;
            
        case VIA_ROTATE_DEGREE_180:             
            RotateDegree = VIA_GEM_ROTATE_180;
            break;
            
        case VIA_ROTATE_DEGREE_270:             
            RotateDegree = VIA_GEM_ROTATE_270;            
            break;                    
    }

    if (pVia->IsHWRotateEnabled)
    {
        /* Modify GEMode and Command according to rotate degree. */
        *pGEMode |= RotateDegree;
        *pCommand |= VIA_GEC_ROT;
    }
}


/* Only some chipsets have direct access window. */
Bool IsSupportDirectAccessWindow(int ChipID)
{
    int IsSupport = FALSE;
    
    switch (ChipID)
    {
        case VIA_PM800:
        case VIA_CX700:
        case VIA_P4M890:
        case VIA_P4M900:
        case VIA_VX800:
            IsSupport = TRUE;
            break;
    }

    return IsSupport;
}


/* Only some chipsets have 2D rotation engine. */
Bool IsSupport2DRotateEngine(int ChipID)
{
    int IsSupport = FALSE;
    
    switch (ChipID)
    {
        case VIA_PM800:
        case VIA_CX700:    
        case VIA_VX800:
            IsSupport = TRUE;
            break;
    }

    return IsSupport;
}


/* If the chipset doesn't have the capabilities to rotate, we have to modify the flag.
*  Note: This function must be called before 2D acceleration initialization. */
void ProbeRotateCaps(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo;
    pBIOSInfo = pVia->pBIOSInfo;    
    
    if (pVia->IsHWRotateEnabled)
    {
        /* Hardware rotation need direct access window, otherwise we can't rotate the screen. */
        if (!IsSupportDirectAccessWindow(pBIOSInfo->Chipset))
        {
            pVia->IsHWRotateEnabled = FALSE;
        }

        /* If the chipset doesn't support rotate engine, we can't turn on 2D acceleration. */
        if (!IsSupport2DRotateEngine(pBIOSInfo->Chipset))
        {
            pVia->NoAccel = TRUE;
        }
    }

}

/* Dynamic rotate screen */
void EnableDynamicRotate(ScrnInfoPtr pScrn)
{    
    VIAPtr  pVia = VIAPTR(pScrn);    
    VIABIOSInfoPtr  pBIOSInfo;
    pBIOSInfo = pVia->pBIOSInfo;
    
    int     dwRotateDegree = pBIOSInfo->HWRotate;    
    CARD32  dwWidth = pBIOSInfo->HDisplay;  /* This value is before rotate. */
    CARD32  dwHeight = pBIOSInfo->VDisplay; /* This value is before rotate. */
    CARD32  dwWidthAlignment8 = 0;
    CARD32  dwWidthAlignment32 = 0;
    
    CARD32 dw1E00 = 0;    /* Degree, Bpp, Enable      */
    CARD32 dw1E04 = 0;    /* start address            */
    CARD32 dw1E08 = 0;    /* end address              */
    CARD32 dw1E0C = 0;    /* source pitch             */
    CARD32 dw1E10 = 0;    /* rotate pitch             */
    CARD32 dw1E14 = 0;    /* W & H                    */    

    if (dwRotateDegree == VIA_ROTATE_DEGREE_0)
    {        
        DisableDirectAccessWindow(pScrn);
        pVia->IsHWRotateEnabled = FALSE;
        return;
    }
    
    /* To do alignment, like 1366x768, 1400x1050... */
    dwWidthAlignment32 = (dwWidth + 31) & (~31);        
    dwWidthAlignment8 = (dwWidth + 7) & (~7);             

    if ((dwWidth == 1400) && (dwHeight == 1050) &&
        (pScrn->bitsPerPixel == 8) &&
        ((dwRotateDegree == VIA_ROTATE_DEGREE_90) || (dwRotateDegree == VIA_ROTATE_DEGREE_270)))
    {
        dwHeight += 2;
    }
    
    /* Set Degree */
    switch (dwRotateDegree)
    {
        case VIA_ROTATE_DEGREE_90:
            dw1E00 |= 0x8;
            break;
            
        case VIA_ROTATE_DEGREE_180:
            dw1E00 |= 0x10;
            break;
                    
        case VIA_ROTATE_DEGREE_270:
            dw1E00 |= 0x18;
            break;
    }

    /* Set Bpp */
    switch (pScrn->bitsPerPixel)
    {
        case 8:
            dw1E00 |= 0x0;
            break;
            
        case 16:
            dw1E00 |= 0x2;
            break;
                    
        case 32:
            dw1E00 |= 0x4;
            break;
    }    
                        
    switch (dwRotateDegree)
    {
        case VIA_ROTATE_DEGREE_90:
            /* 1E08, Set Rotate End Address */
            dw1E08 = dw1E04 + (dwWidthAlignment32 * dwWidthAlignment32 * (pScrn->bitsPerPixel/8));
            /* 1E0C, Rotate Source Pitch */
            dw1E0C = GetReciprocal(dwWidth); 
            /* 1E10, Rotate Pitch */
            dw1E10 = (dwWidthAlignment32 << 16) | dwWidth; 
            /* 1E14, Rotate Height and Width */
            dw1E14 = (dwWidthAlignment8 << 16) | dwWidthAlignment8;
            break;
            
        case VIA_ROTATE_DEGREE_270:            
            /* 1E08, Set Rotate End Address */
            dw1E08 = dw1E04 + (dwWidthAlignment32 * dwWidthAlignment32 * (pScrn->bitsPerPixel/8));
            /* 1E0C, Rotate Source Pitch */
            dw1E0C = GetReciprocal(dwWidthAlignment32); 
            /* 1E10, Rotate Pitch */
            dw1E10 = (dwWidthAlignment32 << 16) | dwWidthAlignment32; 
            /* 1E14, Rotate Height and Width */
            dw1E14 = (dwWidthAlignment8 << 16) | dwHeight;
            break;

        case VIA_ROTATE_DEGREE_0:
        case VIA_ROTATE_DEGREE_180:
            /* 1E08, Set Rotate End Address */
            dw1E08 = dw1E04 + (dwWidthAlignment32 * dwHeight * (pScrn->bitsPerPixel/8));
            dw1E0C = GetReciprocal(dwWidthAlignment32);
            dw1E10 = (dwWidthAlignment32 << 16) | dwWidthAlignment32;
            dw1E14 = (dwHeight << 16) | dwWidthAlignment8;
            break;        
    }        
    
    /* Set Registers */       
    VIASETREG(0x1E04, dw1E04);
    VIASETREG(0x1E08, dw1E08);
    VIASETREG(0x1E0C, dw1E0C);
    VIASETREG(0x1E10, dw1E10);
    VIASETREG(0x1E14, dw1E14);        
    VIASETREG(0x1E00, dw1E00 | 0x1); /* Fire */

}


void ConvertCoordinatesOne(ScrnInfoPtr pScrn, int* x1, int* y1, int w, int h)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo;
    pBIOSInfo = pVia->pBIOSInfo;    
    int newX1 = 0, newY1 = 0;            
    int dwWidth, dwHeight;
    
    switch (pBIOSInfo->HDisplay)
    {
        case 1366:
            dwWidth = pBIOSInfo->HDisplay + 2;
            dwHeight = pBIOSInfo->VDisplay;
            break;        
            
        default:
            dwWidth = pBIOSInfo->HDisplay;
            dwHeight = pBIOSInfo->VDisplay;
            break;
    }

    switch (pBIOSInfo->HWRotate)
    {
        case VIA_ROTATE_DEGREE_0:
            newX1 = *x1;
            newY1 = *y1;                    
            break;
            
        case VIA_ROTATE_DEGREE_90:
            newX1 = dwWidth - *y1 - w;
            newY1 = *x1;                    
            break;
            
        case VIA_ROTATE_DEGREE_180:
            newX1 = dwWidth - *x1 - w;
            newY1 = dwHeight - *y1 - h;                      
            break;
            
        case VIA_ROTATE_DEGREE_270:
            newX1 = *y1;
            newY1 = dwHeight - *x1 - h;                   
            break;
    }
    
    *x1 = newX1;
    *y1 = newY1;    
}

void ConvertCoordinatesTwo(ScrnInfoPtr pScrn, int* x1, int* y1, int* x2, int* y2, int w , int h)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo;
    pBIOSInfo = pVia->pBIOSInfo;              
    int newX1 = 0, newY1 = 0, newX2 = 0, newY2 = 0;                
    int dwWidth, dwHeight;
    
    switch (pBIOSInfo->HDisplay)
    {
        case 1366:
            dwWidth = pBIOSInfo->HDisplay + 2;
            dwHeight = pBIOSInfo->VDisplay;
            break;
            
        default:
            dwWidth = pBIOSInfo->HDisplay;
            dwHeight = pBIOSInfo->VDisplay;
            break;
    }
     
    switch (pBIOSInfo->HWRotate)
    {                    
        case VIA_ROTATE_DEGREE_90:            
            newX1 = dwWidth - *y1 - w;
            newY1 = *x1;        
            newX2 = dwWidth - *y2 - w;
            newY2 = *x2;
            break;
            
        case VIA_ROTATE_DEGREE_180:            
            newX1 = dwWidth - *x1 - w;
            newY1 = dwHeight - *y1 - h;        
            newX2 = dwWidth - *x2 - w;
            newY2 = dwHeight - *y2 - h;
            break;
            
        case VIA_ROTATE_DEGREE_270:            
            newX1 = *y1;
            newY1 = dwHeight - *x1 - h;   
            newX2 = *y2;
            newY2 = dwHeight - *x2 - h;
            break;
    }
    
    *x1 = newX1;
    *y1 = newY1;
    *x2 = newX2;
    *y2 = newY2;    
}


void ConvertDirection(ScrnInfoPtr pScrn, int xdir, int ydir, CARD32* cmd)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo;
    pBIOSInfo = pVia->pBIOSInfo;

    switch (pBIOSInfo->HWRotate)
    {                    
        case VIA_ROTATE_DEGREE_90:
            if (xdir < 0) {
                *cmd |= VIA_GEC_DECY;
            }
            if (ydir > 0) {
                *cmd |= VIA_GEC_DECX;
            }
            break;
            
        case VIA_ROTATE_DEGREE_180:
            if (xdir > 0) {
                *cmd |= VIA_GEC_DECX;
            }
            if (ydir > 0) {
                *cmd |= VIA_GEC_DECY;
            }
            break;
            
        case VIA_ROTATE_DEGREE_270:
            if (xdir > 0) {
                *cmd |= VIA_GEC_DECY;
            }
            if (ydir < 0) {
                *cmd |= VIA_GEC_DECX;
            }
            break;
    }        
}

void ConvertSize(ScrnInfoPtr pScrn, int* width, int* height)
{    
    VIAPtr  pVia = VIAPTR(pScrn);
    VIABIOSInfoPtr  pBIOSInfo;
    pBIOSInfo = pVia->pBIOSInfo;
    int tmpSize;
    
    switch (pBIOSInfo->HWRotate)
    {                    
        case VIA_ROTATE_DEGREE_90:
        case VIA_ROTATE_DEGREE_270:
           tmpSize = *width;
           *width = *height;
           *height = tmpSize;
           break;    
    }
}


 
